use core::arch::global_asm;
use riscv::register::{
    mtvec::TrapMode,
    scause::{self, Exception, Trap},
    stval, stvec
};
use crate::batch::run_next_application;
use crate::println;
use crate::syscall::syscall;
use crate::trap::context::TrapContext;

pub mod context;

global_asm!(include_str!("trap.asm"));

pub fn init() {
    extern "C" { fn __alltraps(); }
    unsafe {
        stvec::write(__alltraps as usize, TrapMode::Direct);
    }
}

#[no_mangle]
pub fn trap_handler(context: &mut TrapContext) -> &mut TrapContext {
    let scause = scause::read();
    let stval = stval::read();

    match scause.cause() {
        Trap::Exception(Exception::UserEnvCall) => {
            context.sepc += 4;
            context.x[10] = syscall(context.x[17], [context.x[10], context.x[11], context.x[12]]);
        },
        Trap::Exception(Exception::StoreFault) | Trap::Exception(Exception::StorePageFault) => {
            println!("[kernel] PageFault in application, kernel will kill it.");
            run_next_application();
        },
        Trap::Exception(Exception::IllegalInstruction) => {
            println!("[kernel] Illegal instruction in application, kernel will kill it.");
            run_next_application();
        },
        _ => {
            panic!("[Kernel] Unsupported trap: {:?}, stval = {:#x}!", scause.cause(), stval);
        }
    }

    context
}